@@ -2,7 +2,7 @@
  * A module for Linux-PAM that will divert to another file and use configuration
  * information from it, percolating the result code back up.  Recursion is fun.
  *
- * Copyright (c) 2000,2001 Red Hat, Inc.
+ * Copyright (c) 2000, 2001, 2004 Red Hat, Inc.
  * Written by Nalin Dahyabhai <nalin@redhat.com>
  * Portions also Copyright (c) 2000 Dmitry V. Levin
  *
@@ -37,20 +37,6 @@
  * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
  * OF THE POSSIBILITY OF SUCH DAMAGE.
  *
- * $Log: pam_stack.c,v $
- * Revision 1.29  2003/07/08 03:16:28  nalin
- * - port to 0.77, mainly using the new pammodutil getXXXXX_r wrappers instead of
- *   the macro forms we used for 0.75
- *
- * Revision 1.28  2001/11/21 19:38:57  nalin
- * free handlers at clean-up time
- *
- * Revision 1.27  2001/11/21 19:38:33  nalin
- * free handlers at clean-up time
- *
- * Revision 1.26  2001/11/21 17:54:59  nalin
- * fix some memory leaks (reported by Fernando Trias)
- *
  */
 
 #define PAM_SM_AUTH
@@ -141,15 +127,15 @@ _pam_stack_copy(pam_handle_t *source, pam_handle_t *dest, unsigned int item,
 			break;
 		case PAM_CONV:
 			name = "PAM_CONV";
-			if(source->pam_conversation && !dest->pam_conversation) {
+			if(source->pam_conversation) {
 				copied = 1;
+				if(dest->pam_conversation) {
+					_pam_drop(dest->pam_conversation);
+				}				
 				dest->pam_conversation = calloc(1, sizeof(struct pam_conv));
 				*dest->pam_conversation = *source->pam_conversation;
 			} else {
-				if(!source->pam_conversation)
-					reason = "source not NULL";
-				if(dest->pam_conversation)
-					reason = "destination already set";
+				reason = "source is NULL";
 			}
 			break;
 		case PAM_FAIL_DELAY:
@@ -283,9 +269,10 @@ _pam_stack_cleanup(pam_handle_t *pamh, void *data, int status)
 		_pam_drop(stack_this->pamh->tty);
 		_pam_drop(stack_this->pamh->rhost);
 		_pam_drop(stack_this->pamh->ruser);
+		_pam_drop(stack_this->pamh->prompt);
 		_pam_drop_env(stack_this->pamh);
 		_pam_drop(stack_this->pamh);
-		free(stack_this->service);
+		_pam_drop(stack_this->service);
 		free(stack_this);
 		stack_this = next;
 	}
@@ -396,12 +383,19 @@ _pam_stack_dispatch(pam_handle_t *pamh, int flags, int argc, const char **argv,
 
 		stack_this = malloc(sizeof(struct stack_data));
 		if(stack_this == NULL) {
+			_pam_drop(service);
 			return PAM_BUF_ERR;
 		}
 
 		memset(stack_this, 0, sizeof(struct stack_data));
-		stack_this->service = _pam_strdup(service);
+		stack_this->service = service;
+
 		stack_this->pamh = calloc(1, sizeof(pam_handle_t));
+		if (stack_this->pamh == NULL) {
+			_pam_drop(stack_this->service);
+			_pam_drop(stack_this);
+			return PAM_BUF_ERR;
+		}
 
 		/* Create an environment for the child. */
 		if(debug) {
@@ -415,16 +409,22 @@ _pam_stack_dispatch(pam_handle_t *pamh, int flags, int argc, const char **argv,
 			syslog(LOG_ERR, "_pam_make_env() returned %s",
 			       pam_strerror(stack_this->pamh, ret));
 			closelog();
+			_pam_drop(stack_this->service);
+			_pam_drop(stack_this->pamh);
+			_pam_drop(stack_this);
 			return PAM_SYSTEM_ERR;
 		}
 
 		/* Set the service.  This loads the service modules. */
-		ret = pam_set_item(stack_this->pamh, PAM_SERVICE, service);
+		ret = pam_set_item(stack_this->pamh, PAM_SERVICE, stack_this->service);
 		if(ret != PAM_SUCCESS) {
 			openlog("pam_stack", LOG_PID, LOG_AUTHPRIV);
 			syslog(LOG_ERR, "pam_set_item(PAM_SERVICE) returned %d (%s)",
 			       ret, pam_strerror(stack_this->pamh, ret));
 			closelog();
+			_pam_drop(stack_this->service);
+			_pam_drop(stack_this->pamh);
+			_pam_drop(stack_this);
 			return PAM_SYSTEM_ERR;
 		}
 
@@ -436,14 +436,22 @@ _pam_stack_dispatch(pam_handle_t *pamh, int flags, int argc, const char **argv,
 			syslog(LOG_ERR, "_pam_init_handlers() returned %d (%s)",
 			       ret, pam_strerror(stack_this->pamh, ret));
 			closelog();
+			_pam_drop(stack_this->service);
+			_pam_drop(stack_this->pamh);
+			_pam_drop(stack_this);
 			return PAM_SYSTEM_ERR;
 		}
 
 		/* Insert the data item at the end of the stack list, or make
 		 * it the head if we don't have one yet. */
 		if(stack_data == NULL) {
-			pam_set_data(pamh, STACK_DATA_NAME, stack_this,
-				     _pam_stack_cleanup);
+			if (pam_set_data(pamh, STACK_DATA_NAME, stack_this,
+				     _pam_stack_cleanup) != PAM_SUCCESS) {
+					_pam_drop(stack_this->service);
+        				_pam_drop(stack_this->pamh);
+                			_pam_drop(stack_this);
+					return PAM_SYSTEM_ERR;
+				}
 		} else {
 			while(stack_data->next != NULL) {
 				stack_data = stack_data->next;
@@ -458,6 +466,7 @@ _pam_stack_dispatch(pam_handle_t *pamh, int flags, int argc, const char **argv,
 			       "`%s'", service);
 			closelog();
 		}
+		_pam_drop(service);
 	}
 	stack_this->debug = debug;
 
@@ -471,7 +480,9 @@ _pam_stack_dispatch(pam_handle_t *pamh, int flags, int argc, const char **argv,
 			closelog();
 		}
 		pam_putenv(stack_this->pamh, env[i]);
+		_pam_drop(env[i]);
 	}
+	_pam_drop(env);
 
 	/* Copy named PAM items to the child. */
 	_pam_stack_copy(pamh, stack_this->pamh, PAM_AUTHTOK, debug ? "child" : NULL);
@@ -517,7 +528,9 @@ _pam_stack_dispatch(pam_handle_t *pamh, int flags, int argc, const char **argv,
 			closelog();
 		}
 		pam_putenv(pamh, env[i]);
+		_pam_drop(env[i]);
 	}
+	_pam_drop(env);
 
 	/* Now the named data items. */
 	_pam_stack_copy(stack_this->pamh, pamh, PAM_AUTHTOK, debug ? "parent" : NULL);
